/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

#ifndef SOCKET_DATA_IO
#define SOCKET_DATA_IO

#include "support/PortableMacros.h"

#ifndef BEOS_OLD_NETSERVER
#include <fcntl.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#endif

#include <sys/time.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include "dataio/PortableDataIO.h"

//: Data I/O to and from a TCP socket! 
class TCPSocketDataIO : public PortableDataIO
{
public:
   // (sockfd) the socket to use
   // blockingIO says whether to use blocking socket I/O or non-blocking
   // socket I/O.  If you will be using this object with a AbstractMessageIOGateway,
   // and/or select(), then it's usually better to set blocking to false.
   TCPSocketDataIO(int sockfd, bool blocking) : _sockfd(sockfd)
   {
      SetBlockingIOEnabled(blocking);
   }

   // Closes the held socket descriptor.
   ~TCPSocketDataIO() {closesocket(_sockfd);}

   // Reads bytes from the socket and places them into (buffer).
   // See PortableDataIO::Read() for semantics.
   virtual ssize_t Read(void * buffer, size_t size)  
   {
      ssize_t ret = recv(_sockfd, buffer, size, 0L);
      return CalculateReturnValue(ret);
   }

   // Reads bytes from (buffer) and sends them out to the TCP socket.
   // See PortableDataIO::Read() for semantics.
   virtual ssize_t Write(const void * buffer, size_t size)
   {
      ssize_t ret = send(_sockfd, buffer, size, 0L);
      return CalculateReturnValue(ret);
   }

   // Flushes the output buffer by turning off Nagle's Algorithm
   // and then turning it back on again
   virtual void FlushOutput()
   {
      SetNaglesAlgorithmEnabled(false);
      SetNaglesAlgorithmEnabled(true);
   }
   
   // Enables or diables blocking I/O on this socket.
   // If this object is to be used by an AbstractMessageIOGateway,
   // then non-blocking I/O is usually better to use.
   status_t SetBlockingIOEnabled(bool blocking)
   {
      long b = blocking ? 0 : 1; 
#ifdef __BEOS__
      if (setsockopt(_sockfd, SOL_SOCKET, SO_NONBLOCK, &b, sizeof(b)) >= 0)
#else
      if (fcntl(_sockfd, F_SETFL, blocking ? 0 : O_NONBLOCK) == 0)
#endif
      {
         _blocking = blocking;
         return B_NO_ERROR;
      }
      else
      {
         perror("TCPSocketDataIO:SetBlockingIO failed");
         return B_ERROR;
      }
   }

   // Turns Nagle's algorithm (output packet buffering/coalescing) on or off.
   status_t SetNaglesAlgorithmEnabled(bool enabled)
   {
#ifdef BEOS_OLD_NETSERVER
      return B_ERROR;  // old networking stack doesn't support this flag
#else
      long delay = enabled ? 0 : 1;
      if (setsockopt(_sockfd, IPPROTO_TCP, TCP_NODELAY, (char *) &delay, sizeof(int)) >= 0)
      {
         return B_NO_ERROR;
      }
      else
      {
         perror("TCPSocketDataIO:SetNaglesAlgorithmEnabled(): setsockopt() failed!");
         return B_ERROR;
      }
#endif
   } 

   // Sets the parameters to reflect the current state of the socket
   // (setReadyForDataRead) - will be set true iff there is incoming data ready to be read.
   // (setReadyForDataWrite) - will be set true iff there is room to send more data.
   // Returns B_NO_ERROR on success, B_ERROR on failure.
   status_t PollSocketReadiness(bool & setReadyForDataRead, bool & setReadyForDataWrite) const
   {
      fd_set readReady, writeReady;
      struct timeval nowait = {0L, 0L};
      FD_ZERO(&readReady);
      FD_ZERO(&writeReady);
      FD_SET(_sockfd,&readReady);
      FD_SET(_sockfd,&writeReady);
      if (select(_sockfd+1, &readReady, &writeReady, NULL, &nowait) >= 0)
      {
         setReadyForDataRead = FD_ISSET(_sockfd,&readReady);
         setReadyForDataWrite = FD_ISSET(_sockfd,&writeReady);
         return B_NO_ERROR;
      }
      return B_ERROR;
   }

   // Releases control of the contained socket to the calling code.
   // After this method returns, this object no longer owns or can
   // use the socket descriptor it once held.
   void ReleaseSocket() {_sockfd = -1;}

   // Returns the socket descriptor held by this object, or
   // -1 if there is none.
   int GetSocket() const {return _sockfd;}

private:
   int CalculateReturnValue(int ret)
   {
      int retForBlocking = ((ret == 0) ? -1 : ret);
      return (_blocking) ? retForBlocking
                         : (((ret < 0)&&(errno == EWOULDBLOCK)) ? 0 : retForBlocking); 
   }

   int _sockfd;
   bool _blocking;
};

#endif
