/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <typeinfo>
#include <string.h>
#include <sys/socket.h>

#ifndef __BEOS__
#include <arpa/inet.h>
#endif

#include <netdb.h>

#include "reflector/ReflectServer.h"
#include "netutil/NetworkUtilityFunctions.h"

status_t
ReflectServer ::
AddNewSession(AbstractReflectSessionRef ref, int socket)
{
   AbstractReflectSession * newSession = ref.GetItemPointer();
   newSession->_socket = socket;
   newSession->_owner = this;  // in case CreateGateway() needs to use the owner

   if (socket >= 0)
   {
      newSession->_gateway = newSession->CreateGateway(socket);
      if (newSession->_gateway == NULL) 
      {
         newSession->_socket = -1;
         return B_ERROR;
      }

      struct sockaddr_in saTempAdd;
#ifdef __BEOS__
      int length = sizeof(sockaddr_in);
#else
      size_t length = sizeof(sockaddr_in);
#endif
      if (getpeername(socket, (struct sockaddr *)&saTempAdd, &length) == 0)
      {
         struct hostent * hePeer = gethostbyaddr((caddr_t)&saTempAdd.sin_addr, sizeof(saTempAdd.sin_addr), AF_INET);

         bool useLocalHostName = false;
         if (hePeer)
         {
            if (strcmp(hePeer->h_name, "localhost")) newSession->_hostName = hePeer->h_name;
                                                else useLocalHostName = true;
         }
         else
         {
             char buf[20];
             Inet_NtoA(ntohl(saTempAdd.sin_addr.s_addr), buf);
             if (strcmp(buf, "127.0.0.1")) newSession->_hostName = buf;
                                      else useLocalHostName = true;
         }
         newSession->_port = ntohs(saTempAdd.sin_port);

         // "localhost" or "127.0.0.1" isn't good enough, we want a REAL hostname!
         if (useLocalHostName)
         {
            char buf[512];
            (void) gethostname(buf, sizeof(buf));
printf("localhost=%s\n", buf);
            newSession->_hostName = buf;
         }
      }
      else newSession->_hostName = "<unknown>";
   }
   else
   {
      newSession->_hostName = "<local>";
      newSession->_port     = 0;
   }

   return AddNewSession(ref);
}


status_t
ReflectServer ::
AddNewSession(AbstractReflectSessionRef ref)
{
   AbstractReflectSession * newSession = ref.GetItemPointer();
   newSession->_owner = this;

   if (_sessions.AddTail(ref))
   {
      _idLookup.Put(newSession->GetSessionIDString(), ref);
      if (newSession->AttachedToServer() == B_NO_ERROR)
      {
         LogTime(LOG_INFO, "New %s [%s] from [%s:%i] (%lu total)\n", newSession->GetTypeName(), newSession->GetSessionIDString(), newSession->GetHostName(), newSession->GetClientPort(), _sessions.CountItems());
         return B_NO_ERROR;
      }
      else 
      {
         LogTime(LOG_ERROR, "%s [%s] from [%s:%i] aborted startup (%lu total)\n", newSession->GetTypeName(), newSession->GetSessionIDString(), newSession->GetHostName(), newSession->GetClientPort(), _sessions.CountItems());
         _sessions.RemoveTail();
         _idLookup.Remove(newSession->GetSessionIDString());
      }
   }

   return B_ERROR;
}


ReflectServer :: ReflectServer() : _idLookup(CStringHashFunc)
{
   _nextPulseAt.tv_sec = _nextPulseAt.tv_usec = 0;

   // This call lets the PortableMessage class use its default message pools for its
   // internal purposes.  It's safe to allow this here, since all PortableMessage calls
   // will be made from the same thread.  For a multi-threaded system, don't do this...
   PortableMessage::SetInternalDefaultPools(GetMessagePool(), GetMessageRefPool());

   _idLookup.SetCompareFunction(CStringCompareFunc);
}

ReflectServer :: ~ReflectServer()
{
   // empty
}

void
ReflectServer :: Cleanup()
{
   _idLookup.Clear();

   // This silly little contortion is necessary to keep the
   // StorageReflectSession destructors from accessing themselves 
   // in the _sessions list while the _sessions list is being cleared!
   // (i.e. we keep an additional reference of each session around in
   // the (temp) array, so that we can clear _sessions without activating
   // all the destructors.  Then when we clear (temp), the destructors are
   // activated, but the _sessions list is already safely cleared)
   PortableQueue<AbstractReflectSessionRef> temp;
   for (int i=_sessions.CountItems()-1; i>=0; i--) temp.AddTail(_sessions[i]);

   _sessions.MakeEmpty();
   _lameDuckSessions.MakeEmpty();
   temp.MakeEmpty();
}

status_t
ReflectServer ::
ReadyToRun()
{
   return B_NO_ERROR;
}

const char *
ReflectServer ::
GetServerName() const
{
   return "MUSCLE";
}

status_t 
ReflectServer ::
ServerProcessLoop(ReflectSessionFactoryFunc createSessionFunc, short acceptPort)
{
   // Setup the connection-accepting socket
   struct sockaddr_in saSocket;
   const long trueValue = 1;
   memset(&saSocket, 0, sizeof(saSocket));
   saSocket.sin_family      = AF_INET;
   saSocket.sin_addr.s_addr = htonl(INADDR_ANY);
   saSocket.sin_port        = htons(acceptPort);

   int acceptSocket;
   if ((acceptSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
   {
      perror("Error creating accept socket!");
      return B_ERROR;
   }
   (void) setsockopt(acceptSocket, SOL_SOCKET, SO_REUSEADDR, &trueValue, sizeof(long));
   if ((bind(acceptSocket, (struct sockaddr *) &saSocket, sizeof(saSocket)))||(listen(acceptSocket, 20)))
   {
      perror("Error binding/listening on accept socket!");
      closesocket(acceptSocket);
      return B_ERROR;
   }

   LogTime(LOG_INFO, "This %s server was compiled on " __DATE__ " " __TIME__ "\n", GetServerName());
   if (ReadyToRun() != B_NO_ERROR) 
   {
      LogTime(LOG_CRITICALERROR, "Server:  ReadyToRun() failed, aborting.\n");
      return B_ERROR;
   }

   LogTime(LOG_INFO, "%s Server is active and listening on port %i\n",GetServerName(), acceptPort);
   fd_set readSet;
   fd_set writeSet;
   while(true)
   {
      // Remove any sessions that were marked for removal
      // (Done first since this may have occurred even in ReadyToRun())
      {
         AbstractReflectSessionRef duckRef;
         while(_lameDuckSessions.RemoveHead(duckRef))
         {
            AbstractReflectSession * duck = duckRef.GetItemPointer();
            for (int j=_sessions.CountItems()-1; j>=0; j--)
            {
               AbstractReflectSession * next = _sessions.GetItemPointer(j)->GetItemPointer();
               if (next == duck)
               {
                  const char * id = next->GetSessionIDString();
                  LogTime(LOG_INFO, "Closed %s [%s] to [%s:%i] (%lu total)\n", next->GetTypeName(), id, next->GetHostName(), next->GetClientPort(), _sessions.CountItems()-1);
                  _idLookup.Remove(id);
                  _sessions.RemoveItemAt(j);
                  break;
               }
            }
         }
      }

      FD_ZERO(&readSet);
      FD_ZERO(&writeSet);
      FD_SET(acceptSocket, &readSet);

      // Find out which of our sessions wants to write data
      int maxSocket = acceptSocket;
      {
         for (int i=_sessions.CountItems()-1; i>=0; i--)
         {
            AbstractReflectSession * session = _sessions.GetItemPointer(i)->GetItemPointer();
            if (session->_gateway)
            {
               int sessionSocket = session->_socket;
               bool in  = session->_gateway->GetInputEnabled(); 
               bool out = session->_gateway->HasBytesToOutput();     
               if (in)  FD_SET(sessionSocket, &readSet);
               if (out) FD_SET(sessionSocket, &writeSet);
               if (((in)||(out))&&(sessionSocket > maxSocket)) maxSocket = sessionSocket;
            }
         }
      }

      // Calculate timeout for Pulse(), if any
      struct timeval waitTime;
      {
         struct timeval now; gettimeofday(&now, NULL);

         GetPulseTime(now, _nextPulseAt);

         // If we are now expecting a Pulse(), calculate how long to wait for it.
         if ((_nextPulseAt.tv_sec != 0)||(_nextPulseAt.tv_usec != 0))
         {
            // estimate how long to wait until the pulse time arrives
            waitTime = _nextPulseAt;
            SubtractTimeVal(waitTime, now);
            const struct timeval zero = {0L, 0L};
            if (IsLessThan(waitTime, zero)) waitTime = zero;
         }
      }
 
      int numBitsSet = select(maxSocket+1, &readSet, &writeSet, NULL, ((_nextPulseAt.tv_sec != 0)||(_nextPulseAt.tv_usec != 0)) ? &waitTime : NULL);
      if (numBitsSet < 0) 
      {
         closesocket(acceptSocket);
         return B_ERROR;
      }

      if ((_nextPulseAt.tv_sec != 0)||(_nextPulseAt.tv_usec != 0))
      {
         struct timeval now; gettimeofday(&now, NULL);
         if (IsLessThan(_nextPulseAt, now))
         {
            Pulse(now);
            _nextPulseAt.tv_sec = _nextPulseAt.tv_usec = 0;
         }
      }

      if (FD_ISSET(acceptSocket, &readSet)) 
      {
         // Accept new connection and start up a session for it
#ifdef __BEOS__
         int nLen = sizeof(saSocket);
#else
         size_t nLen = sizeof(saSocket);
#endif
         int newSocket = accept(acceptSocket, (struct sockaddr *)&saSocket, &nLen);
         if (newSocket >= 0)
         {
            AbstractReflectSession * newSession = createSessionFunc();
            if (newSession) 
            {
               AbstractReflectSessionRef newSessionRef(newSession, NULL, NULL);
               if (AddNewSession(newSessionRef, newSocket) != B_NO_ERROR)
               {
                  perror("Error adding new session!");
                  closesocket(newSocket);
               }
            }
            else
            {
               perror("Error, couldn't create new reflector session!");
               closesocket(newSocket);
            }
         }
         else perror("Warning, accept() failed!");
      }

      // Do the session I/Os...
      {
         for (int i=_sessions.CountItems()-1; i>=0; i--)
         {
            AbstractReflectSession & session = *(_sessions.GetItemPointer(i)->GetItemPointer());
            if (session._gateway)
            {
               AbstractMessageIOGateway & gateway = *(session._gateway);
               int socket = session._socket;
               if (socket >= 0)
               {
                  bool readError  = (FD_ISSET(socket, &readSet))&&(gateway.DoInput()   != B_NO_ERROR);
                  bool writeError = (FD_ISSET(socket, &writeSet))&&(gateway.DoOutput() != B_NO_ERROR); 
                  if ((readError)||(writeError)) _lameDuckSessions.AddTail(_sessions[i]);
               }
            }
         }
      }

      // Let the various sessions handle their messages...
      while(true)
      {
         bool loopAgain = false;

         PortableMessageRef nextMessageRef;
         int numSessions = _sessions.CountItems();
         for (int i=0; ((loopAgain == false)&&(i<numSessions)); i++)
         {
            AbstractReflectSessionRef sRef = _sessions[i];  // used to keep session from deleting itself prematurely
            AbstractReflectSession * s = sRef.GetItemPointer();

            // Note:  It *is* necessary to check fromSession->_gateway every time through this
            //        loop, because a ReplaceSession() call might set it to NULL.
            while((s->_gateway)&&(s->_gateway->GetNextIncomingMessage(nextMessageRef) == B_NO_ERROR)) 
            {
               s->MessageReceivedFromGateway(nextMessageRef);
               if (s->_gateway == NULL)
               {
                  // oops!  This session must have replaced itself.  Start through the loop again.
                  loopAgain = true;
                  break;
               }
            }
         }

         if (loopAgain == false) break;
      }
   }
}

void 
ReflectServer ::
Pulse(const struct timeval & now)
{
   for (int i=_sessions.CountItems()-1; i>=0; i--)
   {
      AbstractReflectSession & session = *(_sessions.GetItemPointer(i)->GetItemPointer());
      if (session._gateway)
      {
         AbstractMessageIOGateway & g = *(session._gateway);
         struct timeval & tv = g._nextPulseAt;
         if (!IsLessThan(now, tv))
         {
            g.Pulse(now);
            tv.tv_sec = tv.tv_usec = 0;
         }
      }
   }
}

void
ReflectServer ::
GetPulseTime(const struct timeval & now, struct timeval & rt)
{
   for (int i=_sessions.CountItems()-1; i>=0; i--)
   {
      AbstractReflectSession & session = *(_sessions.GetItemPointer(i)->GetItemPointer());
      if (session._gateway)
      {
         AbstractMessageIOGateway & g = *(session._gateway);
         struct timeval & tv = g._nextPulseAt;
         if ((tv.tv_sec == 0)&&(tv.tv_usec == 0)) g.GetPulseTime(now, tv);
         if (((tv.tv_sec != 0)||(tv.tv_usec != 0))&&(((rt.tv_sec == 0)&&(rt.tv_usec == 0))||(IsLessThan(tv, rt)))) rt = tv;
      }
   }
}


PortableQueue<AbstractReflectSessionRef> & 
ReflectServer ::
GetSessionsList()
{
   return _sessions;
}


status_t
ReflectServer ::
ReplaceSession(AbstractReflectSessionRef newSessionRef, AbstractReflectSession * oldSession)
{
   // move the gateway from the old session to the new one...
   AbstractReflectSession * newSession = newSessionRef.GetItemPointer();

   newSession->_gateway  = oldSession->_gateway;
   newSession->_hostName = oldSession->_hostName;
   newSession->_port     = oldSession->_port;
   newSession->_socket   = oldSession->_socket;

   if (AddNewSession(newSessionRef) == B_NO_ERROR)
   {
      oldSession->_gateway = NULL;   /* gateway now belongs to newSession */
      EndSession(oldSession);
      return B_NO_ERROR;
   }
   else
   {
       // Oops, rollback changes and error out
       newSession->_gateway  = NULL;
       newSession->_hostName = "";
       newSession->_port     = 0; 
       newSession->_socket   = -1;
       return B_ERROR;
   }
}


void
ReflectServer ::
EndSession(AbstractReflectSession * who)
{
   for (int i=_sessions.CountItems()-1; i>=0; i--)
   {
      if (_sessions.GetItemPointer(i)->GetItemPointer() == who)
      {
         _lameDuckSessions.AddTail(_sessions[i]);
         return;
      }
   }
}
