/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

#ifndef STORAGE_REFLECT_SESSION_H
#define STORAGE_REFLECT_SESSION_H

#include "reflector/DumbReflectSession.h"
#include "reflector/StorageReflectConstants.h"
#include "regex/StringMatcher.h"

//:This is a callback function that may be passed to the ServerProcessLoop() function.
// It creates and returns a new StorageReflector object.
AbstractReflectSession * StorageSessionCreationFunc();

//: A more intelligent AbstractReflectSession that knows how to
//: store data on the server, and filter using wildcards.  This class
//: is used by the muscled server program to handle incoming connections.
//See StorageReflectConstants.h and/or "The Beginner's Guide.html"
//for details.
class StorageReflectSession : public DumbReflectSession
{
public:
   StorageReflectSession();
   ~StorageReflectSession();

   // Called after the constructor, when the session is ready to interact with the server.
   virtual status_t AttachedToServer();

   // Called when a new message is received from our IO gateway.
   virtual void MessageReceivedFromGateway(PortableMessageRef msg);

protected:
   // Create or Set the value of a data node.
   // (nodePath) should be the path relative to the home dir (e.g. "MyNode/Child1/Grandchild2")
   // (dataMsgRef) is the value to set the node to
   // (allowOverwriteData) indicates whether existing node-data may be overwritten.  If false, the method will fail if the specified node already exists.
   // (allowCreateNode) indicates whether new nodes may be created.  (If false,
   // the method will fail if any node in the specified node path 
   // doesn't already exist)
   // Returns B_NO_ERROR on success, or B_ERROR on failure.
   status_t SetDataNode(const PortableString & nodePath, PortableMessageRef dataMsgRef, bool allowOverwriteData=true, bool allowCreateNode=true);

   virtual const char * GetTypeName() const {return "Storage Reflect Session";}

   class DataNode;
   typedef PortableRef<DataNode> DataNodeRef;
   typedef PortableHashtableIterator<const char *, DataNodeRef> DataNodeRefIterator;

   //:Each object of this class represents one node in the server-side data-storage tree.
   class DataNode 
   {
   public:
      void PutChild(DataNodeRef & child, StorageReflectSession * optNotifyWithOnSetParent, StorageReflectSession * optNotifyWithOnChangedData);
      bool HasChild(const char * key) const;
      status_t GetChild(const char * key, DataNodeRef & returnChild) const;
      status_t RemoveChild(const char * key, StorageReflectSession * optNotifyWith, bool recurse);
      DataNodeRefIterator GetChildIterator() const;
      uint32 CountChildren() const;

      const char * GetNodeName() const;
      const char * GetNodePath() const;

      void SetData(PortableMessageRef data, StorageReflectSession * optNotifyWith);
      PortableMessageRef GetData();

      // Calculates and returns our full path (e.g. "/hostname/sessionname/blah/blah/blah/_nodeName"
      DataNode * GetParent() const;
      int GetDepth() const;

      // Returns us to our virgin, pre-Init() state.
      void Reset();  

      // Don't use this, use StorageReflectSession::GetNewDataNode() instead!
      DataNode();

      // Don't use this, use StorageReflectSesssion::ReleaseDataNode() instead!
      ~DataNode();

      // Changes the refcount for the given sessionID.
      // Any sessionID's with (refCount > 0) will be in the GetSubscribers() list.
      void IncrementSubscriptionRefCount(const char * sessionID, long delta);

      // Returns an iterator for our subscribers
      PortableHashtableIterator<const char *, uint32> GetSubscribers() const;

   private:
      friend class StorageReflectSession;

      void Init(const char * nodeName, PortableMessageRef initialValue);
 
      void SetParent(DataNode * _parent, StorageReflectSession * optNotifyWith);

      DataNode * _parent;
      PortableMessageRef _data;
      PortableHashtable<const char *, DataNodeRef> _children;
      PortableString _nodeName;
      PortableString _nodePath;
      int _depth;  // number of ancestors our node has (e.g. root's _depth is zero)

      PortableHashtable<const char *, uint32> _subscribers;
   };

   // This typedef represents the proper signature of a node-tree traversal callback function.
   // Functions with this signature may be used with the WildPathMatcher::DoTraversal() method.
   // (This) is the StorageReflectSession doing the traveral, as specified in the DoTraversal() call
   // (node) is the DataNode that was matched the criteria of the traversal
   // (userData) is the same value that was passed in to the DoTraversal() method.
   // The callback function should return the depth at which the traversal should continue after
   // the callback is done.  So to allow the traversal to continue normally, return (node.GetDepth()),
   // or to terminate the traversal immediately, return 0, or to resume the search at the next
   // session, return 2, or etc.
   typedef int (*PathMatchCallback)(StorageReflectSession * This, DataNode & node, void * userData);
  
   typedef PortableRef<StringMatcher> StringMatcherRef;
   typedef PortableRef<PortableQueue<StringMatcherRef> > StringMatcherQueueRef;

   //:This class is used to do efficient pattern-matching on wild paths (e.g. "/*/*/jeremy*/fries*")
   class WildPathMatcher
   {
   public:
      WildPathMatcher();
      ~WildPathMatcher();

      void Clear();

      status_t AddPathString(const PortableString & wildpath);
      void AddPathsFromMatcher(const WildPathMatcher & matcher);
      status_t AddPathsFromMessage(const char * fieldName, const PortableMessage & msg, const char * optPrependIfNoLeadingSlash);
      status_t RemovePathString(const PortableString & wildpath);

      // Returns true iff the given node matches our query.
      bool MatchesNode(DataNode & node) const;

      // Returns true iff the given path string is in our string set
      bool ContainsPathString(const PortableString & w) const;

      // Does a depth-first traversal of the node tree, starting with (node) as the root.
      // (cb) is the callback function to call whenever a node is encountered in the traversal
      //      that matches at least one of our path strings.
      // (node) is the node to start the traversal at.
      // (userData) is any value you wish; it will be passed along to the callback method.
      void DoTraversal(PathMatchCallback cb, StorageReflectSession * This, DataNode & node, void * userData);
 
      // If (w) starts with '/', this method will remove the slash.
      // If it doesn't start with a '/', and (optPrepend) is non-NULL,
      // this method will prepend (optPrepend+"/") to (w).
      void AdjustStringPrefix(PortableString & w, const char * optPrepend) const;

      // Returns the number of path-strings that we contain that match (node).
      // Note this is a bit more expensive than MatchesNode(), as we can't use
      // short-circuit boolean logic here!
      int GetMatchCount(DataNode & node) const;

   private:
      int DoTraversalAux(PathMatchCallback cb, StorageReflectSession * This, DataNode & node, void * userData, int rootDepth);
      bool PathMatches(DataNode & node, int whichPath) const;
      PortableQueue<StringMatcherQueueRef> _parsers;
      PortableQueue<PortableString> _pathStrings;
   };

   friend class StorageReflectSession::DataNode;
   friend void ResetNodeFunc(DataNode *, void *);

   // Removes our nodes from the tree;  called by the destructor or on setup errors.
   void Cleanup();

   // Tells other sessions that we have modified (node).
   void NotifySubscribersThatNodeChanged(DataNode & node, bool isBeingRemoved);

   // Tells other sessions that we have a new (node) available.
   void NotifySubscribersOfNewNode(DataNode & newNode);

   // Called by NotifySubscribersOfNewNode to tell us that (node) has been modified.
   void NodeChanged(DataNode & node, bool isBeingRemoved);

   // Called by SetParent to tell us that (node) has been created at a given location.
   // We then respond by letting any matching subscriptions add their mark to the node.
   void NodeCreated(DataNode & node);

   // Takes any messages that were created in the NodeChanged() callbacks and 
   // sends them to their owner's MessageReceivedFromNeighbor() method for
   // processing and eventual forwarding to the client.
   void PushSubscriptionMessages();

   // Executes a data-gathering tree traversal based on the PR_NAME_KEYS
   // specified in the given message.
   void DoGetData(const PortableMessage & getMsg);

   // If set false, we won't receive subscription updates
   void SetSubscriptionsEnabled(bool e) {_subscriptionsEnabled = e;}

   // Accessor
   bool GetSubscriptionsEnabled() const {return _subscriptionsEnabled;}

   // Convenience method:  Uses the given path to lookup a single node in the node tree
   // and return it.  (Note that wildcards are not supported by this method!)
   // If (path) begins with a '/', the search will begin with the root
   // node of the tree; if not, it will begin with this session's node.  Returns NULL on failure.
   DataNode * GetDataNode(const PortableString & path);

   // Traversal callback:  matching nodes are ref'd/unref'd with subscribed session IDs
   int DoSubscribeRefCallback(DataNode & node, void * userData);

   // callback stub
   static int DoSubscribeRefCallbackFunc(StorageReflectSession * This, DataNode & node, void * userData) 
   {
       return This->DoSubscribeRefCallback(node, userData);
   }

   // Traversal callback:  matching nodes are sent the given message (userData).
   int PassMessageCallback(DataNode & node, void * userData);

   // callback stub
   static int PassMessageCallbackFunc(StorageReflectSession * This, DataNode & node, void * userData) 
   {
       return This->PassMessageCallback(node, userData);
   }

   // Traversal callback:  matching nodes are placed into the given message (userData).
   int GetDataCallback(DataNode & node, void * userData);

   // callback stub
   static int GetDataCallbackFunc(StorageReflectSession * This, DataNode & node, void * userData) 
   {
       return This->GetDataCallback(node, userData);
   }

   // Traversal callback:  matching nodes are placed in a list (userData) to be
   // removed from the node tree.
   int RemoveDataCallback(DataNode & node, void * userData);
   static int RemoveDataCallbackFunc(StorageReflectSession * This, DataNode & node, void * userData) 
   {
       return This->RemoveDataCallback(node, userData);
   }

   // Call this to get a new DataNode, instead of using the DataNode ctor directly
   DataNode * GetNewDataNode(const char * nodeName, PortableMessageRef initialValue);

   // Call this when you are done with a DataNode, instead of the DataNode destructor
   void ReleaseDataNode(DataNode * node);

   PortableMessage _parameters;  // our parameter set
   PortableMessage * _nextSubscriptionMessage;  // will be sent when a subscription triggers
   DataNode * _globalRoot;       // Points to shared root for all sessions
   DataNodeRef _sessionDir;      // this session's subdir (grandchild of _globalRoot)
   WildPathMatcher _subscriptions;
   WildPathMatcher _defaultMessageRoute;
   bool _subscriptionsEnabled;

   // Below:  item pools to reduce memory allocations/deallocations

   static DataNodeRef::ItemPool _nodePool;
   static DataNodeRef::RefPool _nodeRefPool;

   static StringMatcherRef::ItemPool _stringMatcherPool;
   static StringMatcherRef::RefPool _stringMatcherRefPool;

   static StringMatcherQueueRef::ItemPool _stringMatcherQueuePool;
   static StringMatcherQueueRef::RefPool _stringMatcherQueueRefPool;

   friend class StorageReflectSession :: WildPathMatcher;
};

#endif
