/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

/******************************************************************************
/
/	File:		PortableFlattenable.h
/
/	Description:	Portable version of Be's BFlattenable base class.
/
******************************************************************************/

#ifndef	PORTABLE_FLATTENABLE_H
#define	PORTABLE_FLATTENABLE_H

#include <string.h>
#include "support/PortableTypeDefs.h"
#include "support/PortableStatusConstants.h"

/*--------------------------------------------------------------------*/
/*----- PortableFlattenable class ------------------------------------*/

//: This class is an interface representing an object that knows how
//: to save itself into an array of bytes, and recover its state from
//: an array of bytes.
class PortableFlattenable 
{
public:
   // Should return true iff every object of this type has a size that is known at compile time.
   virtual	bool		IsFixedSize() const = 0;

   // Should return the type code identifying this type of object.
   virtual	type_code	TypeCode() const = 0;

   // Should return the number of bytes needed to store this object in its current state.
   virtual	size_t		FlattenedSize() const = 0;

   // Should store this object's state into (buffer).  (buffer) must be
   // (FlattenedSize()) bytes long.
   virtual	void		Flatten(uint8 *buffer) const = 0;

   // Should return true iff a buffer with type_code (code) can be used to reconstruct
   // this object's state.  Defaults to returning true iff (code) equals TypeCode().
   virtual	bool		AllowsTypeCode(type_code code) const {return (code == TypeCode());}

   // Should attempt to restore this object's state from the given buffer.
   virtual	status_t	Unflatten(const uint8 *buf, size_t size) = 0;

   virtual			~PortableFlattenable() {}

   // Convenience methods for flattening and unflattening, with proper bounds checking
   // and so on to avoid crashes due to bad data.  Return true iff successful.

   // Writes data consecutively into a byte buffer.  The output buffer is
   // assumed to be large enough to hold the data.
   // - outBuf Flat buffer to write to
   // - writeOffset Offset into buffer to read from.  Incremented by (blockSize) on success.
   // - copyFrom memory location to copy bytes from
   // - blockSize number of bytes to copy
   static void WriteData(uint8 * outBuf, size_t * writeOffset, const void * copyFrom, size_t blockSize)
   {
      memcpy(&outBuf[*writeOffset], copyFrom, blockSize);
      *writeOffset += blockSize;
   };
    
   // Reads data safely from a byte buffer
   // - inBuf Flat buffer to read bytes from
   // - outputBufferBytes total size of the input buffer
   // - readOffset Offset into buffer to read from.  Incremented by (blockSize) on success.
   // - copyTo memory location to copy bytes to
   // - blockSize number of bytes to copy
   static bool ReadData(const uint8 * inBuf, size_t inputBufferBytes, size_t * readOffset, void * copyTo, size_t blockSize)
   {
      if ((*readOffset + blockSize) > inputBufferBytes) return false;
      memcpy(copyTo, &inBuf[*readOffset], blockSize);
      *readOffset += blockSize;
      return true;
   };
};

/*-------------------------------------------------------------*/
/*-------------------------------------------------------------*/

#endif /* _PORTABLEFLATTENABLE_H */

