/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

/* Note:  This file was originally written by Michael Olivero, but has been modified
 *        extensively.  Nonetheless, the original license is listed below.  --jaf
 */

/*
   This library was downloaded from: http://www.mike95.com

   This library is copyright.  It may freely be used for personal purposes 
   if the restriction listed below is adhered to.
       Author: Michael Olivero
       Email:  mike95@mike95.com

   //===============================
   //Start of Restriction Definition
   //===============================
   Anyone can have full use of the library provided they keep this complete comment
   with the source.  Also I would like to ask if any changes are made to the
   code for efficiency reasons, please let me know so I may look into your change and
   likewise incorporate it into library.  If a suggestion makes it into the library,
   your credits will be added to this information.

   Authors of Computer related books are welcome to include this source code as part
   of their publishing, provided credit the Author and make note of where the source
   code was obtained from: http://www.mike95.com
   //=============================
   //End of Restriction Definition
   //=============================

   Description:
   Visit http://www.mike95.com/c_plusplus/classes/JHashtable/

   This class is a based on the Java Hashtable class and as such contains all the public
   member functions of its Java equivalent.  Unlike Java, typecasts are not necessary
   since C++ allows template instatiation of types at compile time.

   Note:  Since java has hashCode() as a member of the base Object class, all 
   Java classes are inherently hashable.  Since the template parameter types do
   not necessarily have to have a built in hashing function, the user of the class
   must specify a hash function by calling setHashFunction() passing a pointer 
   to the hash function.

   The has function must be declared as the following:

       uint32 function(const KeyType&);

   Where:
   function = any name you choose to use for the function name
   KeyType  = the type used for the key in the construction of the PortableHashtable object.

   Example:
       uint32 myHash( const int&)
       {
           //your hashing code here for a key of type int.
       }
*/

#ifndef _PORTABLE_HASHTABLE_H
#define _PORTABLE_HASHTABLE_H

#include <new.h>
#include "support/PortableTypeDefs.h"
#include "support/PortableMacros.h"

template <class KeyType, class ValueType> class PortableHashtable;  // forward declaration

//:This class is an iterator object, used for iterating over the set
//:of keys or values in a PortableHashtable. 
// Note that it is illegal to modify the hashtable during an iteration 
// traversal (an assertion failure will occur if you try it, and your 
// program will exit)
//
// Given a PortableHashtable object, you can obtain one or more of these
// iterator objects by calling the PortableHashtable's GetIterator() method.
//
// This iterator actually contains separate state for two iterations:
// one for iterating over the values in the PortableHashtable, and one for 
// iterating over the keys.  These two iterations can be done independently
// of each other.
template <class KeyType, class ValueType> class PortableHashtableIterator
{
public:
   // Default constructor.  It's here only so that you can include PortableHashtableIterators
   // as member variables, in arrays, etc.  PortableHashtableIterators created with this
   // constructor are "empty", so they won't be useful until you set them equal to a 
   // PortableHashtableIterator that was returned by PortableHashtable::GetIterator().
   PortableHashtableIterator();

   // Returns true iff there are more keys left in the key traversal.
   bool HasMoreKeys() const;

   // If there are more keys left in the key traversal, copies the next one into
   // (setNextKey) and returns true.  Otherwise, returns false.
   bool GetNextKey(KeyType & setNextKey);

   // As above, but returns a pointer to the key object, or NULL if none are left.
   // (Be careful, the returned pointer may become invalid when the hashtable is modified)
   KeyType * GetNextKey();

   // Copies the next key in the key traversal into (setKey) without modifying
   // the state of the traversal.  Returns false iff there are no more keys left.
   bool PeekNextKey(KeyType & setKey) const;

   // As above, but returns a pointer to the key object, or NULL if none left.
   // (Be careful, the returned pointer may become invalid when the hashtable is modified)
   KeyType * PeekNextKey() const;

   // Returns true iff there are more values left in the value traversal.
   bool HasMoreValues() const;

   // If there are more values left in the values traversal, copies the next one into
   // (setNextValue) and returns true.  Otherwise, returns false.
   bool GetNextValue(ValueType & setNextValue);

   // As above, but returns a pointer to the next value object, or NULL if none are left.
   // (Be careful, the returned pointer may become invalid when the hashtable is modified)
   ValueType * GetNextValue();

   // Copies the next value in the value traversal into (setValue) without modifying
   // the state of the traversal.  Returns false iff there are no more values left.
   bool PeekNextValue(ValueType & setKey) const;

   // As above, but returns a pointer to the next value object, or NULL if none are left.
   // (Be careful, the returned pointer may become invalid when the hashtable is modified)
   ValueType * PeekNextValue() const;

   // (note that the default copy constructor and '=' operator are okay for our purposes,
   //  as we don't "own" any of the data that we point to)

private:
   friend class PortableHashtable<KeyType, ValueType>;

   PortableHashtableIterator(const PortableHashtable<KeyType, ValueType> & owner, uint32 stateCode);

   void * _nextKeyCookie;
   int32 _nextKeyIndex;

   void * _nextValueCookie;
   int32 _nextValueIndex;

   uint32 _stateCode;  // used to detect modifications-during-iterator-traversals and trigger assertion failures

   const PortableHashtable<KeyType, ValueType> * _owner;
};


//:This is a handy templated Hashtable class, similar to Java's java.util.Hashtable.
template <class KeyType, class ValueType> class PortableHashtable
{
public:
   // (func) specified a function that returns an integer hashcode for a given key object.
   // (initialCapacity) specifies the number of slots to initially pre-allocate.   Defaults to 13.
   // (loadFactor) specifies how densely populated the table must be before it is re-hashed.  Defaults to 0.5 (i.e. 50%)
   PortableHashtable(uint32(*func)(const KeyType& key), uint32 initialCapacity = 13, float loadFactor = 0.5f);
   virtual ~PortableHashtable();

   //Inspectors
   //==========

   // Returns the number of items in the table
   uint32 Size() const {return _count;}

   // Returns true iff the table is empty
   bool IsEmpty() const {return _count == 0;}

   // Returns true iff the table contains a mapping with the given value.  (O(n) search time)
   bool Contains(const ValueType& value) const;

   // Returns true iff the table contains a mapping with the given key.  (O(1) search time)
   bool ContainsKey( const KeyType& key) const;

   // If the given (key) is in the table, copies the associated value into (setValue)
   // and returns true.  Otherwise returns false.
   bool Get(const KeyType& key, ValueType& setValue) const; 

   // As above, only returns a pointer to the internally held copy of the value
   // object.  This is more efficient than Get() but a little more risky, as the
   // returned pointer may become invalid the next time this hashtable is modified.
   // So be careful!  Returns NULL if the given key isn't in the table.
   ValueType * Get(const KeyType & key) const;

   // Returns an iterator object that can be used to enumerate the items in the hash table.
   PortableHashtableIterator<KeyType,ValueType> GetIterator() const 
     {return PortableHashtableIterator<KeyType,ValueType>(*this, _stateCode);}

   //Modifiers
   //=========

   // Places the given (key, value) mapping into the table.  Any previous entry with
   // a key of (key) will be replaced.  If there was a previously existing mapping
   // with the given (key), its value will be copied into (setPreviousValue) and this
   // method will return true.  Otherwise, this method will return false.
   bool Put(const KeyType& key, const ValueType& value, ValueType & setPreviousValue);

   // Places the given (key, value) mapping into the table.  Any previous entry with
   // a key of (key) will be replaced.
   void Put(const KeyType& key, const ValueType& value);

   // Removes the mapping with the given (key).  Returns true on success, false on
   // failure (i.e. no mapping with the given (key) was found)
   bool Remove(const KeyType& key);

   // Removes the mapping with the given (key) and places its value into (setRemovedValue).
   // Returns true on success, false on failure (i.e. no mapping with the given (key) was found).
   bool Remove(const KeyType& key, ValueType & setRemovedValue);

   // Removes all mappings from the hash table.
   void Clear();

   // You can call this to change the hash function that is used to generate integer
   // hashing values from the key items.
   void SetHashFunction(uint32(*func)(const KeyType& key)) {_userHashFunc = func;}

   // You can call this to set a function to use to compare keys.  By default, the 
   // '==' operator is used, but for some key types (e.g. const char *'s) '==' is not
   // useful, so you can supply your own here.
   void SetCompareFunction(bool(*func)(const KeyType& key1, const KeyType& key2)) {_userCompareFunc = func;}

private:
   friend class PortableHashtableIterator<KeyType, ValueType>;

   PortableHashtable(const PortableHashtable &);  // unimplemented and inaccessible
   PortableHashtable& operator=(const PortableHashtable &msg); // ditto

   class PortableHashtableEntry
   {
   public:
      PortableHashtableEntry() : _valid(false), _next(NULL) {/* empty */}

      // Returns the entry to its just-created state
      void Invalidate()
      {
         KeyType defaultKey;     _key   = defaultKey;
         ValueType defaultValue; _value = defaultValue;
         _valid = false;
         _next = NULL;
         // hash is undefined, so no point in resetting it here
      }

      uint32 _hash;                   // precalculated for efficiency
      KeyType _key;                   // used for '==' checking
      ValueType _value;               // payload
      PortableHashtableEntry* _next;  // for making linked lists
      bool _valid;                    // only used for entries in our _table array, (which are by value and so can't be NULL)
   };

   // Auxilliary methods
   bool PutAux(const KeyType& key, const ValueType& value, ValueType * optSetPreviousValue);
   bool RemoveAux(const KeyType& key, ValueType * setRemovedValue);

   void GetNextCookie(void ** c, int32 * setIndex, uint32 stateCode) const 
   {
      ASSERT(stateCode==_stateCode, "Illegal operation detected:  PortableHashtable modified during traversal!");

      PortableHashtableEntry * cookie = *((PortableHashtableEntry **) c);
      if (cookie) *c = cookie = cookie->_next;  // no "else" to this if!
      if (cookie == NULL)
      {
         // Gotta go through the table and find the next entry!
         if (_table)
         {
            int32 index = (*setIndex)+1;
            while((index >= 0)&&(index < _tableSize))
            {
               PortableHashtableEntry * e = &_table[index];
               if (e->_valid)
               {
                  *c = e;
                  *setIndex = index;
                  return;
               }
               else index++;
            }
         }

         // If we got here, then we're all out of entries
         *setIndex = _tableSize;
         *c = NULL;
      }
   }

   KeyType * GetKeyFromCookie(void * c) const {return &(((PortableHashtableEntry *)c)->_key);}
   ValueType * GetValueFromCookie(void * c) const {return &(((PortableHashtableEntry *)c)->_value);}

   bool KeysAreEqual(const KeyType & key1, const KeyType & key2) const;
   uint32 NextPrime(uint32 start);
   void GrowTable();

   uint32 _count;       //the size of the elements in the hashtable
   uint32 _tableSize;   //the size of the table.
   uint32 _threshold;
   float _loadFactor;
   PortableHashtableEntry * _table;
   uint32 (*_userHashFunc)(const KeyType& key);
   bool (*_userCompareFunc)(const KeyType& key1, const KeyType& key2);
   uint32 _stateCode;  // used to detect modifications-during-iterator-traversals
};

//===============================================================
// Implementation of PortableHashtable
// Necessary location for appropriate template instantiation.
//===============================================================
template <class KeyType, class ValueType>
PortableHashtable<KeyType,ValueType>::PortableHashtable(uint32 (*func)(const KeyType & key), uint32 initialCapacity, float loadFactor)
   : _userHashFunc(func), _userCompareFunc(NULL), _stateCode(0), _count(0), _tableSize(initialCapacity), _loadFactor(loadFactor), _threshold((uint32)(initialCapacity*loadFactor)), _table(NULL)
{
   // empty
}

template <class KeyType, class ValueType>
PortableHashtable<KeyType,ValueType>::~PortableHashtable()
{
   Clear();
   delete [] _table;
}

template <class KeyType, class ValueType>
bool 
PortableHashtable<KeyType,ValueType>::KeysAreEqual(const KeyType & key1, const KeyType & key2) const
{
   return (_userCompareFunc) ? _userCompareFunc(key1, key2) : (key1 == key2);
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::Contains(const ValueType& value) const
{
   if (_table) 
   {
      for(uint32 i = 0; i < _tableSize; i++)
      {
         PortableHashtableEntry * e = &_table[i];
         if (e->_valid)
         {
            while(e)
            {
               if (e->_value == value) return true;
               e = e->_next;
            }
         }
      }
   }
   return false;
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::ContainsKey(const KeyType& key) const
{
   return (Get(key) != NULL);
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::Get(const KeyType& key, ValueType & setValue) const
{
   ValueType * ptr = Get(key);
   if (ptr)
   {
      setValue = *ptr;
      return true;
   }
   else return false;
}


template <class KeyType, class ValueType>
ValueType *
PortableHashtable<KeyType,ValueType>::Get(const KeyType& key) const
{
   if (_table)
   {
      uint32 hash = (*_userHashFunc)(key);
      PortableHashtableEntry * e = &_table[hash % _tableSize];
      if (e->_valid)
      {
         while(e)
         {
            if ((e->_hash == hash)&&(KeysAreEqual(e->_key, key))) return &e->_value;
            e = e->_next; 
         }
      }
   }
   return NULL;
}

template <class KeyType, class ValueType>
void
PortableHashtable<KeyType,ValueType>::GrowTable()
{
   _stateCode++;

   if (_table)
   {
      uint32 oldTableSize = _tableSize;
      PortableHashtableEntry * oldTable = _table;

      uint32 newTableSize = NextPrime(2 * oldTableSize);
      PortableHashtableEntry * newTable = new (nothrow) PortableHashtableEntry[newTableSize];
      if (newTable == NULL) {WARN_OUT_OF_MEMORY; return;}

      _threshold = (uint32)(newTableSize * _loadFactor);
      _table = newTable;

      // move all entries from the old array to the new one
      for(uint32 i = 0; i < oldTableSize; i++)
      {
         PortableHashtableEntry * anchor = &oldTable[i];
         if (anchor->_valid)
         {
            PortableHashtableEntry * e = anchor;
            while(e)
            {
               PortableHashtableEntry * next = e->_next;
               PortableHashtableEntry * newSlot = &_table[e->_hash % newTableSize];
               if (newSlot->_valid)
               {
                  if (e == anchor) e = new (nothrow) PortableHashtableEntry(*anchor);
                  if (e)
                  {
                     PortableHashtableEntry * oldSecond = newSlot->_next;
                     newSlot->_next = e;
                     e->_next = oldSecond;
                  }
                  else WARN_OUT_OF_MEMORY;  // trouble!
               }
               else
               {
                  *newSlot = *e;
                  newSlot->_next = NULL;
                  if (e != anchor) delete e;
               }
               e = next;
            }
         }
      }
      delete [] oldTable;
      _tableSize = newTableSize;
   }
}

template <class KeyType, class ValueType>
void
PortableHashtable<KeyType,ValueType>::Put(const KeyType& key, const ValueType& value)
{
   (void) PutAux(key, value, NULL);  
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::Put(const KeyType& key, const ValueType& value, ValueType & setPreviousValue)
{
   return PutAux(key, value, &setPreviousValue);
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::PutAux(const KeyType& key, const ValueType& value, ValueType * optSetPreviousValue)
{
   _stateCode++;

   if (_table == NULL)  // demand-allocate this
   {
      _table = new (nothrow) PortableHashtableEntry[_tableSize];
      if (_table == NULL) {WARN_OUT_OF_MEMORY; return false;}
   }

   uint32 hash = (*_userHashFunc)(key);
   uint32 index = hash % _tableSize;

   //Makes sure the key is not already in the hashtable
   PortableHashtableEntry * e = &_table[index];
   if (e->_valid)
   {
      while(e)
      {
         if ((e->_hash == hash)&&(KeysAreEqual(e->_key, key))) 
         {
            if (optSetPreviousValue) *optSetPreviousValue = e->_value;
            e->_value = value;
            return true;
         }
         e = e->_next;
      }
   }

   // Rehash the table if the threshold is exceeded
   if (_count >= _threshold)
   {
      GrowTable();
      return PutAux(key, value, optSetPreviousValue);
   }

   PortableHashtableEntry * slot = &_table[index];
   if (slot->_valid)
   {
      // Creates new entry in linked list for this slot
      PortableHashtableEntry * newEntry = new (nothrow) PortableHashtableEntry();
      if (newEntry == NULL) {WARN_OUT_OF_MEMORY; return false;}
      newEntry->_hash  = hash;
      newEntry->_key   = key;
      newEntry->_value = value;
      newEntry->_valid = true;
      newEntry->_next  = slot->_next;
      slot->_next = newEntry;
   }
   else
   {
      // First entry in slot; just copy data over
      slot->_hash  = hash;
      slot->_key   = key;
      slot->_value = value;  
      slot->_valid = true;
   }
   _count++;

   return false;
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::Remove(const KeyType& key)
{
   return RemoveAux(key, NULL);
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::Remove(const KeyType& key, ValueType & setValue)
{
   return RemoveAux(key, &setValue);
}


template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::RemoveAux(const KeyType& key, ValueType * optSetValue)
{
   _stateCode++;
   if (_table)
   {
      uint32 hash = (*_userHashFunc)(key);
      PortableHashtableEntry * e = &_table[hash % _tableSize];
      if (e->_valid)
      {
         PortableHashtableEntry * prev = NULL;
         while(e)
         {
            if ((e->_hash == hash)&&(KeysAreEqual(e->_key, key)))
            {
               if (optSetValue) *optSetValue = e->_value;
               PortableHashtableEntry * next = e->_next;
               if (prev)
               {
                  prev->_next = next;
                  delete e;
               }
               else
               {
                  if (next) 
                  {
                     *e = *next;
                     delete next;
                  }
                  else e->Invalidate();
               }
               _count--;
               return true;
            }
            prev = e;
            e = e->_next;
         }
      }
   }
   return false;
}

template <class KeyType, class ValueType>
void
PortableHashtable<KeyType,ValueType>::Clear()
{
   _stateCode++;
   if (_table)
   {
      for(uint32 i = 0; i < _tableSize; i++)
      {
         PortableHashtableEntry * anchor = &_table[i];
         if (anchor->_valid)
         {
            PortableHashtableEntry * e = anchor->_next;
            while(e)
            {
               PortableHashtableEntry * temp = e;
               e = e->_next;
               delete temp;
            }
            anchor->Invalidate();
         }
      }
   }
   _count = 0;
}

template <class KeyType, class ValueType>
uint32
PortableHashtable<KeyType,ValueType>::NextPrime(uint32 start)
{
   if (start % 2 == 0) start++;
   uint32 i;
   for(; ; start += 2)
   {
      for(i = 3; i * i <= start; i += 2) if (start % i == 0) break;
      if (i * i > start) return start;
   }
}

//===============================================================
// Implementation of PortableHashtableIterator
//===============================================================

template <class KeyType, class ValueType>
PortableHashtableIterator<KeyType, ValueType>::PortableHashtableIterator() : _owner(NULL), _nextKeyCookie(NULL), _nextKeyIndex(-1), _stateCode(0)
{
   // empty
}

template <class KeyType, class ValueType>
PortableHashtableIterator<KeyType, ValueType>::PortableHashtableIterator(const PortableHashtable<KeyType, ValueType> & owner, uint32 stateCode) : _owner(&owner), _nextKeyCookie(NULL), _nextKeyIndex(-1), _stateCode(stateCode)
{
   _owner->GetNextCookie(&_nextKeyCookie, &_nextKeyIndex, _stateCode);

   // Initially, the two traversal states both point to the first item
   _nextValueCookie = _nextKeyCookie;
   _nextValueIndex  = _nextKeyIndex;
}

template <class KeyType, class ValueType>
bool
PortableHashtableIterator<KeyType,ValueType>::HasMoreKeys() const
{
   return (_nextKeyCookie != NULL);
}

template <class KeyType, class ValueType>
bool
PortableHashtableIterator<KeyType,ValueType>::HasMoreValues() const
{
   return (_nextValueCookie != NULL);
}

template <class KeyType, class ValueType>
bool
PortableHashtableIterator<KeyType,ValueType>::GetNextKey(KeyType& key) 
{
   KeyType * ret = GetNextKey();
   if (ret) 
   {
      key = *ret;
      return true;
   }
   else return false;
}

template <class KeyType, class ValueType>
bool
PortableHashtableIterator<KeyType,ValueType>::GetNextValue(ValueType& val)
{
   ValueType * ret = GetNextValue();
   if (ret)
   {
      val = *ret;
      return true;
   } 
   else return false;
}

template <class KeyType, class ValueType>
bool
PortableHashtableIterator<KeyType,ValueType>::PeekNextKey(KeyType& key) const
{
   KeyType * ret = PeekNextKey();
   if (ret)
   {
      key = *ret;
      return true;
   }
   else return false;
}

template <class KeyType, class ValueType>
bool
PortableHashtableIterator<KeyType,ValueType>::PeekNextValue(ValueType& val) const
{
   ValueType * ret = PeekNextValue();
   if (ret)
   {
      val = *ret;
      return true;
   }
   else return false;
}

template <class KeyType, class ValueType>
ValueType *
PortableHashtableIterator<KeyType,ValueType>::GetNextValue()
{
   if (_owner)
   {
      ValueType * val = PeekNextValue();
      if (val)
      {
         _owner->GetNextCookie(&_nextValueCookie, &_nextValueIndex, _stateCode);
         return val;
      }
   }
   return NULL;
}

template <class KeyType, class ValueType>
KeyType *
PortableHashtableIterator<KeyType,ValueType>::GetNextKey() 
{
   if (_owner)
   {
      KeyType * ret = PeekNextKey();
      if (ret)
      {
         _owner->GetNextCookie(&_nextKeyCookie, &_nextKeyIndex, _stateCode);
         return ret;
      }
   }
   return NULL;
}

template <class KeyType, class ValueType>
ValueType *
PortableHashtableIterator<KeyType,ValueType>::PeekNextValue() const
{
   return ((_owner)&&(_nextValueCookie)) ? _owner->GetValueFromCookie(_nextValueCookie) : NULL;
}

template <class KeyType, class ValueType>
KeyType *
PortableHashtableIterator<KeyType,ValueType>::PeekNextKey() const
{
   return ((_owner)&&(_nextKeyCookie)) ? _owner->GetKeyFromCookie(_nextKeyCookie) : NULL;
}

#endif

